# src/settings.py

import os

# lexi-system variables
COMPOSE_PROJECT_NAME = os.getenv("COMPOSE_PROJECT_NAME")

# Confluence environment variables
CONFLUENCE_PRIVATE_API_KEY = os.getenv("CONFLUENCE_PRIVATE_API_KEY")
CONFLUENCE_SPACE_URL = os.getenv("CONFLUENCE_SPACE_URL")
CONFLUENCE_SPACE_KEY = os.getenv("CONFLUENCE_SPACE_KEY")
CONFLUENCE_EMAIL_ADDRESS = os.getenv("CONFLUENCE_EMAIL_ADDRESS")
CONFLUENCE_SPACE_NAME = os.getenv("CONFLUENCE_SPACE_NAME")

ENVRC_TEMPLATE_FILE = os.getenv("ENVRC_TEMPLATE_FILE")



from abc import ABC

import tiktoken
from tqdm import tqdm
from wasabi import msg

from lexi.verba_interface.chunk import Chunk
from lexi.verba_interface.document import Document


class Chunker(ABC):
    """
    Interface for Verba Chunking.
    """

    def __init__(self):
        self.default_units = 100
        self.default_overlap = 50

    def chunk(
        self, documents: list[Document], units: int, overlap: int
    ) -> list[Document]:
        """Chunk verba documents into chunks based on units and overlap.

        @parameter: documents : list[Document] - List of Verba documents
        @parameter: units : int - How many units per chunk (words, sentences, etc.)
        @parameter: overlap : int - How much overlap between the chunks
        @returns list[str] - List of documents that contain the chunks.
        """
        raise NotImplementedError("chunk method must be implemented by a subclass.")


class TokenChunker(Chunker):
    """
    TokenChunker for Verba built with tiktoken.
    """

    def __init__(self):
        super().__init__()
        self.name = "TokenChunker"
        self.requires_library = ["tiktoken"]
        self.default_units = 250
        self.default_overlap = 50
        self.description = "Chunk documents by tokens powered by tiktoken. You can specify how many tokens should overlap between chunks to improve retrieval."
        self.encoding = tiktoken.encoding_for_model("gpt-3.5-turbo")

    def chunk(
        self, documents: list[Document], units: int, overlap: int
    ) -> list[Document]:
        """Chunk verba documents into chunks based on units and overlap
        @parameter: documents : list[Document] - List of Verba documents
        @parameter: units : int - How many units per chunk (words, sentences, etc.)
        @parameter: overlap : int - How much overlap between the chunks
        @returns list[str] - List of documents that contain the chunks.
        """
        for document in tqdm(
            documents, total=len(documents), desc="Chunking documents"
        ):
            # Skip if document already contains chunks
            if len(document.chunks) > 0:
                continue

            encoded_tokens = self.encoding.encode(document.text, disallowed_special=())

            if units > len(encoded_tokens) or units < 1:
                doc_chunk = Chunk(
                    text=document.text,
                    doc_name=document.name,
                    doc_type=document.type,
                    chunk_id=0,
                )

            if overlap >= units:
                msg.warn(
                    f"Overlap value is greater than unit (Units {units}/ Overlap {overlap})"
                )
                continue

            i = 0
            split_id_counter = 0
            while i < len(encoded_tokens):
                # Overlap
                start_i = i
                end_i = min(i + units, len(encoded_tokens))

                chunk_tokens = encoded_tokens[start_i:end_i]
                chunk_text = self.encoding.decode(chunk_tokens)

                doc_chunk = Chunk(
                    text=chunk_text,
                    doc_name=document.name,
                    doc_type=document.type,
                    chunk_id=split_id_counter,
                )
                document.chunks.append(doc_chunk)
                split_id_counter += 1

                # Exit loop if this was the last possible chunk
                if end_i == len(encoded_tokens):
                    break

                i += units - overlap  # Step forward, considering overlap

        return documents



from typing import List

from langchain_community.document_loaders import ConfluenceLoader
from settings import (CONFLUENCE_EMAIL_ADDRESS, CONFLUENCE_PRIVATE_API_KEY,
                      CONFLUENCE_SPACE_URL)
from verba_interface.document import Document as VerbaDocument

from .document_importer import DocumentImporter


class ConfluenceImporter(DocumentImporter):
    def __init__(self):
        self.loader = ConfluenceLoader(
            url=CONFLUENCE_SPACE_URL,
            username=CONFLUENCE_EMAIL_ADDRESS,
            api_key=CONFLUENCE_PRIVATE_API_KEY,
        )

    def import_documents(self, space_key: str, limit: int = 50) -> List[VerbaDocument]:
        documents = self.loader.load(
            space_key=space_key, include_attachments=False, limit=limit
        )
        # Convert loaded documents to VerbaDocuments here
        return [self._convert_to_verba_document(doc) for doc in documents]

    def _convert_to_verba_document(self, document) -> VerbaDocument:
        # Conversion logic
        pass



from typing import List

from verba_interface.document import Document as VerbaDocument


class DocumentProcessor:
    def process_documents(self, documents: List[VerbaDocument]) -> List[VerbaDocument]:
        raise NotImplementedError



from lexi.verba_interface.chunk import Chunk


class Document:
    def __init__(
        self,
        text: str = "",
        type: str = "",
        name: str = "",
        path: str = "",
        link: str = "",
        timestamp: str = "",
        reader: str = "",
        meta: dict = None,
    ):
        if meta is None:
            meta = {}
        self._text = text
        self._type = type
        self._name = name
        self._path = path
        self._link = link
        self._timestamp = timestamp
        self._reader = reader
        self._meta = meta
        self.chunks: list[Chunk] = []

    @property
    def text(self):
        return self._text

    @property
    def type(self):
        return self._type

    @property
    def name(self):
        return self._name

    @property
    def path(self):
        return self._path

    @property
    def link(self):
        return self._link

    @property
    def timestamp(self):
        return self._timestamp

    @property
    def reader(self):
        return self._reader

    @property
    def meta(self):
        return self._meta

    @staticmethod
    def to_json(document) -> dict:
        """Convert the Document object to a JSON dict."""
        doc_dict = {
            "text": document.text,
            "type": document.type,
            "name": document.name,
            "path": document.path,
            "link": document.link,
            "timestamp": document.timestamp,
            "reader": document.reader,
            "meta": document.meta,
            "chunks": [chunk.to_dict() for chunk in document.chunks],
        }
        return doc_dict

    @staticmethod
    def from_json(doc_dict: dict):
        """Convert a JSON string to a Document object."""
        document = Document(
            text=doc_dict.get("text", ""),
            type=doc_dict.get("type", ""),
            name=doc_dict.get("name", ""),
            path=doc_dict.get("path", ""),
            link=doc_dict.get("link", ""),
            timestamp=doc_dict.get("timestamp", ""),
            reader=doc_dict.get("reader", ""),
            meta=doc_dict.get("meta", {}),
        )
        # Assuming Chunk has a from_dict method
        document.chunks = [
            Chunk.from_dict(chunk_data) for chunk_data in doc_dict.get("chunks", [])
        ]
        return document



import logging
from logging.handlers import RotatingFileHandler

from rich.console import Console
from rich.theme import Theme


class RichLogger:
    def __init__(self, log_file_path: str = None):
        # Define custom theme
        custom_theme = Theme(
            {"info": "dim cyan", "success": "bold green", "error": "bold red", "warning": "bold yellow"}
        )
        self.console = Console(theme=custom_theme)

        # Set up file logging if a path is provided
        if log_file_path:
            self.file_logger = logging.getLogger("rich_logger")
            self.file_logger.setLevel(logging.DEBUG)  # Or adjust to your needs
            file_handler = RotatingFileHandler(
                log_file_path, maxBytes=1048576, backupCount=5
            )
            log_formatter = logging.Formatter(
                "%(asctime)s - %(levelname)s - %(message)s"
            )
            file_handler.setFormatter(log_formatter)
            self.file_logger.addHandler(file_handler)
        else:
            self.file_logger = None

    def log(self, message: str, level: str = "info"):
        # Print to console with Rich
        self.console.print(message, style=level)

        # Log to file if file logging is set up
        if self.file_logger:
            if level == "info":
                self.file_logger.info(message)
            elif level == "success":
                self.file_logger.info(
                    message
                )  # Logging libraries typically don't have a 'success' level
            elif level == "error":
                self.file_logger.error(message)
            elif level == "warning":
                self.file_logger.warning(message)

    def log_info(self, message: str):
        self.log(message, level="info")

    def log_success(self, message: str):
        self.log(message, level="success")

    def log_error(self, message: str):
        self.log(message, level="error")

    def log_warning(self, message: str):
        self.log(message, level="warning")



import os
import re
import time
from datetime import datetime

import typer
import weaviate
from langchain_community.document_loaders import ConfluenceLoader
from rich.console import Console
from rich.progress import Progress
from rich.table import Table
from tqdm import tqdm
from weaviate.classes.config import DataType, Property

import lexi.settings as settings
from lexi.tui.system import system_app
from lexi.utils.rich_logger import RichLogger
from lexi.verba_interface.chunker import TokenChunker
from lexi.verba_interface.document import Document as VerbaDocument

logger = RichLogger("application.log")

app = typer.Typer()

schema_app = typer.Typer()

app.add_typer(schema_app, name="schema")
app.add_typer(system_app, name="system")



def strip_non_letters(s: str):
    return re.sub(r"[^a-zA-Z0-9]", "_", s)


def import_data(
    documents: list[VerbaDocument],
) -> bool:
    """Import verba documents and its chunks to Weaviate
    @parameter: documents : list[Document] - List of Verba documents
    @parameter: client : Client - Weaviate Client
    @parameter: batch_size : int - Batch Size of Input
    @returns bool - Bool whether the embedding what successful.
    """
    client = _get_weaviate_client()

    # vectorizer = "text2vec-openai"

    Document_collection = client.collections.get("Document_MiniLM")
    Chunk_collection = client.collections.get("Chunk_MiniLM")

    for i, document in enumerate(documents):
        batches = []
        uuid = ""
        temp_batch = []
        token_counter = 0
        for chunk in document.chunks:
            print(chunk.to_dict())
            print(chunk.tokens)
            if token_counter + chunk.tokens <= 4000:
                token_counter += chunk.tokens
                temp_batch.append(chunk)
            else:
                batches.append(temp_batch.copy())
                token_counter = chunk.tokens
                temp_batch = [chunk]
        if len(temp_batch) > 0:
            batches.append(temp_batch.copy())
            token_counter = 0
            temp_batch = []

        typer.echo(
            f"({i+1}/{len(documents)}) Importing document {document.name} with {len(batches)} batches"
        )

        with Document_collection.batch.dynamic() as batch:
            batch.batch_size = 1
            properties = {
                "text": str(document.text),
                "doc_name": str(document.name),
                "doc_type": str(document.type),
                "doc_link": str(document.link),
                "chunk_count": len(document.chunks),
                "timestamp": str(document.timestamp),
            }

            uuid = batch.add_object(properties)

            for chunk in document.chunks:
                chunk.set_uuid(uuid)

        chunk_count = 0
        for _batch_id, chunk_batch in tqdm(
            enumerate(batches), total=len(batches), desc="Importing batches"
        ):
            with Chunk_collection.batch.dynamic() as batch:
                batch.batch_size = len(chunk_batch)
                for i, chunk in enumerate(chunk_batch):
                    chunk_count += 1

                    properties = {
                        "text": chunk.text,
                        "doc_name": str(document.name),
                        "doc_uuid": chunk.doc_uuid,
                        "doc_type": chunk.doc_type,
                        "chunk_id": chunk.chunk_id,
                    }

                    # Check if vector already exists
                    if chunk.vector is None:
                        batch.add_object(properties=properties)
                    else:
                        batch.add_object(properties=properties, vector=chunk.vector)

                    wait_time_ms = int(
                        os.getenv("WAIT_TIME_BETWEEN_INGESTION_QUERIES_MS", "0")
                    )
                    if wait_time_ms > 0:
                        time.sleep(float(wait_time_ms) / 1000)

    return True


def load_data(
    documents,
):
    """
    Load data into the system.
    """
    # Convert each document into VerbaDocument
    verba_documents = []
    for document in documents:
        verba_document = VerbaDocument(
            text=document.page_content,
            type="Documentation",
            name=document.metadata.get("title", ""),
            link=document.metadata.get("source", ""),
            timestamp=str(datetime.now().strftime("%Y-%m-%d %H:%M:%S")),
            reader="bespoke",
        )
        verba_documents.append(verba_document)

    if verba_documents is None:
        typer.echo("No documents imported")
        return

    chunker = TokenChunker()

    print("Before chunking:")
    for doc in verba_documents:
        print(f"Document Name: {doc.name}, Text Length: {len(doc.text)}")

    modified_documents = chunker.chunk(verba_documents, units=250, overlap=50)

    print("After chunking:")
    for doc in modified_documents:
        print(f"Document Name: {doc.name}, Num Chunks: {len(doc.chunks)}")

    import_data(modified_documents)

    document_count = len(modified_documents)
    chunks_count = sum([len(document.chunks) for document in modified_documents])

    typer.echo(
        f"Succesfully imported {document_count} documents and {chunks_count} chunks"
    )


def _get_weaviate_client():
    # Connect with default parameters for a local Weaviate instance
    client = weaviate.connect_to_local()
    # client = weaviate.Client("http://localhost:8080")
    return client


def download_confluence_documents():
    # Initialize the ConfluenceLoader with the settings from settings.py
    loader = ConfluenceLoader(
        url=settings.CONFLUENCE_SPACE_URL,
        username=settings.CONFLUENCE_EMAIL_ADDRESS,  # Assuming the email address is used as the username
        api_key=settings.CONFLUENCE_PRIVATE_API_KEY,
    )

    # Download documents
    documents = loader.load(
        space_key=settings.CONFLUENCE_SPACE_KEY, include_attachments=False, limit=50
    )

    return documents


def import_documents_to_weaviate(documents):
    client = (
        weaviate.connect_to_local()
    )  # Ensure this matches your Weaviate instance configuration

    collection = client.collections.get("Documents")

    with Progress() as progress, collection.batch.dynamic() as batch:
        task = progress.add_task("[cyan]Importing documents...", total=len(documents))

        for document in documents:
            # Assuming 'document' is a dict with the correct structure for your schema
            batch.add_object(
                properties=document.dict(),
                # uuid=document.get("uuid", weaviate.util.generate_uuid5(document))  # Generate UUID for each document
            )
            progress.advance(task)


@schema_app.command("create")
def create_schema(schema_name: str = "Documents"):
    client = _get_weaviate_client()

    try:
        collection = client.collections.get(schema_name)
        collection_config = collection.config.get()
        if collection_config:
            typer.echo(f"Collection '{schema_name}' already exists.")
            return
    except Exception:
        typer.echo(f"Collection '{schema_name}' does not exist, creating.")

    try:
        client.collections.create(
            "Documents",
            properties=[
                Property(name="page_content", data_type=DataType.TEXT),
                Property(name="metadata", data_type=DataType.TEXT),
            ],
        )

        typer.echo(f"Schema '{schema_name}' created successfully.")
    except Exception as e:
        typer.echo(f"Failed to check or create schema: {e}")


@schema_app.command("details")
def schema_details(schema_name: str):
    client = _get_weaviate_client()
    try:
        collection = client.collections.get(schema_name)
        collection_config = collection.config.get()
    except Exception:
        typer.echo(f"Collection '{schema_name}' does not exist.")
        return

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Collection Name", style="cyan")
    table.add_column("Property Name", style="magenta", justify="right")
    table.add_column("Data Type", style="green", justify="right")

    for prop in collection_config.properties:
        table.add_row(
            schema_name,
            prop.name,
            prop.data_type.value,  # Adjust based on actual attribute names
        )

    console = Console()
    console.print(table)


@schema_app.command("list")
def list_schema():
    client = _get_weaviate_client()
    try:
        response = client.collections.list_all()

        table = Table(show_header=True, header_style="bold magenta")
        table.add_column("Collection Name", style="cyan")
        table.add_column("Property Name", style="magenta", justify="right")
        table.add_column("Data Type", style="green", justify="right")

        last_collection_name = ""
        for collection_name, collection_details in response.items():
            for prop in collection_details.properties:
                display_name = (
                    collection_name if collection_name != last_collection_name else ""
                )
                table.add_row(
                    display_name,
                    prop.name,
                    prop.data_type.value,  # Adjust based on actual attribute names
                )
                last_collection_name = collection_name

        console = Console()
        console.print(table)

    except Exception as e:
        typer.echo(f"Failed to list schemas: {e}")


@app.command("load")
def load_confluence_data():
    documents = download_confluence_documents()
    typer.echo(f"Downloaded {len(documents)} documents.")
    # Import documents to Weaviate
    load_data(documents)
    typer.echo(f"Imported {len(documents)} documents into Weaviate.")


@schema_app.command("top_documents")
def list_top_documents(schema_name: str):
    client = _get_weaviate_client()
    try:
        # Query to retrieve top 5 documents based on your criteria, e.g., most recent
        collection = client.collections.get(schema_name)

        table = Table(show_header=True, header_style="bold magenta")
        table.add_column("Document Name", style="cyan")
        table.add_column("Link", style="magenta", justify="right")
        # Add more columns as needed based on the properties of your documents

        for item in collection.iterator():
            table.add_row(item.properties["doc_name"], item.properties["doc_link"])

        console = Console()
        console.print(table)
    except Exception as e:
        typer.echo(f"Failed to list top documents for '{schema_name}': {str(e)}")


if __name__ == "__main__":
    app()



class Chunk:
    def __init__(
        self,
        text: str = "",
        doc_name: str = "",
        doc_type: str = "",
        doc_uuid: str = "",
        chunk_id: str = "",
    ):
        self._text = text
        self._doc_name = doc_name
        self._doc_type = doc_type
        self._doc_uuid = doc_uuid
        self._chunk_id = chunk_id
        self._tokens = 0
        self._vector = None
        self._score = 0

    @property
    def text(self):
        return self._text

    @property
    def text_no_overlap(self):
        return self._text_no_overlap

    @property
    def doc_name(self):
        return self._doc_name

    @property
    def doc_type(self):
        return self._doc_type

    @property
    def doc_uuid(self):
        return self._doc_uuid

    @property
    def chunk_id(self):
        return self._chunk_id

    @property
    def tokens(self):
        return self._tokens

    @property
    def vector(self):
        return self._vector

    @property
    def score(self):
        return self._score

    def set_uuid(self, uuid):
        self._doc_uuid = uuid

    def set_tokens(self, token):
        self._tokens = token

    def set_vector(self, vector):
        self._vector = vector

    def set_score(self, score):
        self._score = score

    def to_dict(self) -> dict:
        """Convert the Chunk object to a dictionary."""
        return {
            "text": self.text,
            "doc_name": self.doc_name,
            "doc_type": self.doc_type,
            "doc_uuid": self.doc_uuid,
            "chunk_id": self.chunk_id,
            "tokens": self.tokens,
            "vector": self.vector,
            "score": self.score,
        }

    @classmethod
    def from_dict(cls, data: dict):
        """Construct a Chunk object from a dictionary."""
        chunk = cls(
            text=data.get("text", ""),
            doc_name=data.get("doc_name", ""),
            doc_type=data.get("doc_type", ""),
            doc_uuid=data.get("doc_uuid", ""),
            chunk_id=data.get("chunk_id", ""),
        )
        chunk.set_tokens(data.get("tokens", 0))
        chunk.set_vector(data.get("vector", None))
        chunk.set_score(data.get("score", 0))
        return chunk



import os
import subprocess
from getpass import getpass
from string import Template

import typer
from typer import Option

import lexi.settings as settings
from lexi.utils.rich_logger import RichLogger

system_folder_path = os.path.join(os.path.dirname(__file__), "../system")

logger = RichLogger("application.log")

system_app = typer.Typer()

@system_app.command("up")
def system_up(
    fail_on_no_openai_api_key: bool = Option(
        False,
        "--fail-on-no-openai-api-key",
        help="Fail if there is no OPENAI_API_KEY in the environment."
    )
):
    """
    Starts the Docker Compose services.
    """
    openai_api_key = os.getenv("OPENAI_API_KEY")
    if not openai_api_key:
        message = "OPENAI_API_KEY is not set in the environment."
        if fail_on_no_openai_api_key:
            logger.log_error(message)
            raise typer.Exit(code=1)
        else:
            logger.log_warning(message)

    try:
        env = os.environ.copy()
        env["COMPOSE_PROJECT_NAME"] = settings.COMPOSE_PROJECT_NAME

        subprocess.run(
            ["docker-compose", "up", "-d"], check=True, cwd=system_folder_path, env=env
        )
        logger.log_success("Docker Compose services have been started.")
    except subprocess.CalledProcessError as e:
        logger.log_error(f"Failed to start Docker Compose services: {e}")


@system_app.command("down")
def system_down():
    """
    Stops the Docker Compose services.
    """
    try:
        subprocess.run(["docker-compose", "down"], check=True, cwd=system_folder_path)
        logger.log_success("Docker Compose services have been stopped.")
    except subprocess.CalledProcessError as e:
        logger.log_error(f"Failed to stop Docker Compose services: {e}")


def load_template(file_path: str) -> Template:
    with open(file_path, 'r', encoding='utf-8') as file:
        template_content = file.read()
        return Template(template_content)

@system_app.command("create_envrc")
def create_envrc(
    confluence_url: str = typer.Option(..., prompt=True),
    confluence_space_key: str = typer.Option(..., prompt=True),
    confluence_email: str = typer.Option(..., prompt=True),
    confluence_space_name: str = typer.Option("", prompt=True),
    compose_project_name: str = typer.Option(..., prompt=True),
    litellm_log: str = typer.Option("INFO", prompt=True),
    force: bool = Option(False, "--force", "-f", help="Force overwrite of existing .envrc file")
):
    """
    Creates a .envrc file with the specified environment variables.
    """
    envrc_file_path = '.envrc'

    # Check if .envrc file exists
    if os.path.exists(envrc_file_path) and not force:
        typer.echo("The .envrc file already exists. Use --force to overwrite.")
        raise typer.Exit()

    # Load the template from the file
    template = load_template(settings.ENVRC_TEMPLATE_FILE)

    confluence_api_key = getpass("Enter Confluence API Key: ")
    openai_api_key = getpass("Enter OpenAI API Key: ")

    # Substitute values in the template
    variables = {
        "confluence_private_api_key": confluence_api_key,
        "confluence_space_url": confluence_url,  # Make sure confluence_url is defined
        "confluence_space_key": confluence_space_key,  # Make sure confluence_space_key is defined
        "confluence_email_address": confluence_email,  # Make sure confluence_email is defined
        "confluence_space_name": confluence_space_name or "",
        "openai_api_key": openai_api_key,
        "compose_project_name": compose_project_name,  # Make sure compose_project_name is defined
        "litellm_log": litellm_log  # Make sure litellm_log is defined
    }

    # Substitute the variables in the template
    envrc_content = template.substitute(variables)


    with open('.envrc', 'w', encoding='utf-8') as f:
        f.write(envrc_content)

    logger.log_success("The .envrc file has been created successfully.")

if __name__ == "__main__":
    system_app()



from typing import List

from verba_interface.document import Document as VerbaDocument


class DocumentImporter:
    def import_documents(self, documents: List[VerbaDocument]) -> None:
        raise NotImplementedError
